/*
 * xforward.c
 * 
 * Program that acts as a x relay
 * 
 * Misc. fixes and Athena widget support by Samuel T. Denton, III
 *
 * C preprocessor controls:
 *   MUST_ALLOW  -- if defined, allowable host names must be specified
 *                  at runtime; otherwise absence of such names allows
 *                  any host to attempt to connect
 *   USE_MOTIF_GUI -- if not defined, use Athena widget set
 *   TRACE_PACKETS -- if defined, trace certain events relating to
 *                    TCP/IP i/o buffering.
 */

/* Copyright 1994 Samuel T. Denton, III
 *                        All Rights Reserved
 *
 * Permission to use, copy, and modify this software and its documentation is
 * hereby granted only under the following terms and conditions.  Both the
 * above copyright notice and this permission notice must appear in all copies
 * of the software, derivative works or modified versions, and any portions
 * thereof, and both notices must appear in supporting documentation.
 *
 * Users of this software agree to the terms and conditions set forth herein,
 * and hereby grant back to the author a non-exclusive, unrestricted,
 * royalty-free right and license under any changes, enhancements or
 * extensions made to the core functions of the software, including but not
 * limited to those affording compatibility with other hardware or software
 * environments, but excluding applications which incorporate this software.
 * Users further agree to use their best efforts to return to the author any
 * such changes, enhancements or extensions that they make and inform the
 * author of noteworthy uses of this software.  Correspondence should be
 * provided to the author at:
 * 
 *                      Samuel T. Denton, III
 *                      400 Harrison Avenue
 *                      Ferguson, MO 63135  
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND SAMUEL T. DENTON, III, DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS.   IN NO EVENT SHALL SAMUEL T. DENTON, III,
 * BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */

/* Copyright 1992, 1993 Digital Equipment Corporation
 *                        All Rights Reserved
 *
 * Permission to use, copy, and modify this software and its documentation is
 * hereby granted only under the following terms and conditions.  Both the
 * above copyright notice and this permission notice must appear in all copies
 * of the software, derivative works or modified versions, and any portions
 * thereof, and both notices must appear in supporting documentation.
 *
 * Users of this software agree to the terms and conditions set forth herein,
 * and hereby grant back to Digital a non-exclusive, unrestricted, royalty-free
 * right and license under any changes, enhancements or extensions made to the
 * core functions of the software, including but not limited to those affording
 * compatibility with other hardware or software environments, but excluding
 * applications which incorporate this software.  Users further agree to use
 * their best efforts to return to Digital any such changes, enhancements or
 * extensions that they make and inform Digital of noteworthy uses of this
 * software.  Correspondence should be provided to Digital at:
 *
 *                      Director of Licensing
 *                      Cambridge Research Laboratory
 *                      Digital Equipment Corporation
 *                      One Kendall Square, Bld. 700
 *                      Cambridge, MA 02139  
 *
 * This software may be distributed (but not offered for sale or transferred 
 * for compensation except on systems manufactured by Digital Equipment 
 * Corporation) to third parties, provided such third parties agree to abide 
 * by the terms and conditions of this notice.  
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND DIGITAL EQUIPMENT CORP. DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS.   IN NO EVENT SHALL DIGITAL EQUIPMENT
 * CORPORATION BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <netdb.h>
#include <signal.h>
#include <errno.h>
#include <sys/param.h>

extern int      errno;

#define min(a,b) ((a) < (b) ? (a) : (b))
#define max(a,b) ((a) > (b) ? (a) : (b))

#define TRUE 1
#define FALSE 0

/* define numbers for lucky and unlucky connections */
#define OK     7
#define CANCEL 13

#define X11_TCP_BASE_PORT 6000

void usage();
void copyfromto();
void NotePipeError();
void CheckDisplay();
void CreatePopup();
void alert_user_CB();

struct pending {
	pid_t child;
	int newsock;
	struct pending *next;
};

struct iobuffer {
	int bytes; /* nbr of bytes waiting in buffer */
	int opposite; /* fd this one should be connected to */
	int wclose; /* if set, close after finished writing */
	char buffer[BUFSIZ + 8];
};

struct iobuffer *iob;

fd_set rinit, winit;
int nfds;

char *fwd2name = NULL;
struct sockaddr_in fwd2addr;

char *progname;

main(argc, argv)
	int argc;
	char *argv[];
{
	struct hostent *hp;
	char myhostname[MAXHOSTNAMELEN];
	struct sockaddr_in local;
	unsigned int   *okhosts = NULL;
	char **okhost_names = NULL;
	unsigned int    nokhosts = 0;
	unsigned short  port;
	int i;
	int lsock;
	int one = 1;
	fd_set readable, writable;
	int dtsize;
	char *hname;
	size_t len;
	pid_t child_pid;
	struct pending *pend_head = NULL;
	int tcount = 0;
	int terse = 0;

	progname = argv[0];
	if ((argc == 1) && (*argv[0] == '-')) {
		char *cp = malloc(BUFSIZ);

#ifndef	_POSIX_ARG_MAX
#define	_POSIX_ARG_MAX 4096
#endif
		argc = 0;
		argv = (char**)malloc(_POSIX_ARG_MAX * sizeof(char*));
		gets(cp);
		cp = strtok(cp, " \t");
		do {
			argv[argc++] = cp;
			cp = strtok(NULL, " \t");
		} while (cp);
		argv[argc] = NULL;
		terse = 1; /* force terse mode */
	} else {
		/* Set default from environment */
		fwd2name = getenv("DISPLAY");
	}

	/* Check parameters */
#ifdef MUST_ALLOW
	if (argc < 3)
		usage();
#endif
	for (i = 1; i < argc; i++) {
		if (strcmp(argv[i], "-display") == 0) {
			fwd2name = argv[++i];
		} else if (strcmp(argv[i], "-terse") == 0) {
			terse = 1;
		} else if (strcmp(argv[i], "-allow") == 0) {
			if (!okhosts) {
				int nhosts = argc - i - 1;
				okhosts = (unsigned int *) calloc(nhosts, sizeof *okhosts);
				okhost_names = (char **) calloc(nhosts, sizeof *okhost_names);
			}
			while (argv[++i] && argv[i][0] != '-') {
				if ((hp = gethostbyname(argv[i])) == NULL) {
					fprintf(stderr, "no such host %s\n", argv[i]);
					exit(1);
				}
				okhost_names[nokhosts] = malloc(strlen(argv[i]) + 1);
				strcpy(okhost_names[nokhosts], argv[i]);
				bcopy(hp->h_addr, (char *) &okhosts[nokhosts], 4);
				nokhosts++;
			}
			i--;	/* cancel out impending for-loop increment */
		} else {
			usage();
		}
	}
	/* Error if not set */
	if (fwd2name == NULL) {
		fprintf(stderr, "destination display not specified\n");
		exit(1);
	}
#ifdef MUST_ALLOW
	if (nokhosts == 0) {
		fprintf(stderr, "must specify at least one allowed host.\n");
		exit(1);
	}
#endif
	/* Get my (firewall's) hostname, and make sure it's usable. */
	gethostname(myhostname, MAXHOSTNAMELEN);
	if ((hp = gethostbyname(myhostname)) == NULL) {
		fprintf(stderr, "Can't lookup up my own hostname %s\n", myhostname);
		exit(1);
	}
	strcpy(myhostname, hp->h_name);

	/* Check that we can open the destination */
	CheckDisplay(fwd2name);

	/* get the destination's hostname, and make sure it's usable. */
	len = strcspn(fwd2name, ":");
	hname = malloc(len + 1);/* allocate space, including null */
	(void) strncpy(hname, fwd2name, len);
	hname[len] = '\0';
	hp = gethostbyname(hname);
	if (!hp) {
		fprintf(stderr, "no such host %s\n", hname);
		exit(1);
	}
	port = X11_TCP_BASE_PORT + atoi(fwd2name + len + 1);
	if (port < X11_TCP_BASE_PORT || port > (X11_TCP_BASE_PORT + 100)) {
		fprintf(stderr, "display number must be >= 0 and <= 100\n");
		exit(1);
	}
	bzero(&fwd2addr, sizeof(fwd2addr));
	fwd2addr.sin_family = AF_INET;
	fwd2addr.sin_port = htons(port);
	bcopy(hp->h_addr, &fwd2addr.sin_addr, sizeof(fwd2addr.sin_addr));


	/* find/create socket for local "display" */
	if ((lsock = socket(PF_INET, SOCK_STREAM, 0)) < 0) {
		perror("socket");
		exit(1);
	}
	if (setsockopt(lsock, SOL_SOCKET, SO_REUSEADDR,
		 &one, sizeof(one)) < 0) {
		perror("SO_REUSEADDR:");
		exit(1);
	}
	bzero(&local, sizeof(local));
	local.sin_family = AF_INET;
	local.sin_port = htons(port = X11_TCP_BASE_PORT);
	while (bind(lsock, (struct sockaddr *) & local, sizeof(local)) < 0) {
		if (errno == EADDRINUSE) {
			port++;
			if (port > X11_TCP_BASE_PORT + 100) {
				fprintf(stderr, "all ports in use.\n");
				exit(1);
			}
			local.sin_port = htons(port);
		} else {
			perror("bind");
			exit(1);
		}
	}

	if (listen(lsock, SOMAXCONN) < 0) {
		perror("listen");
		exit(1);
	}
	if (!terse)
		printf("display is ");
	printf("%s:%d\n", myhostname, port - X11_TCP_BASE_PORT);
	fflush(stdout);

	signal(SIGPIPE, SIG_IGN);

	dtsize = getdtablesize();

	/* set up iobuffers & sizes */
	iob = (struct iobuffer *) calloc(dtsize, sizeof *iob);
	if (!iob) {
		fprintf(stderr, "can't allocate iobuffers\n");
		exit(1);
	}
	/* -1 means not connected/invalid */
	while (dtsize--) {
		iob[dtsize].opposite = -1;
	}

	FD_ZERO(&rinit);
	FD_ZERO(&winit);
	FD_SET(0, &rinit);
	FD_SET(lsock, &rinit);

	nfds = lsock + 1;
	while (1) {
		struct timeval  timeout;
		int nready;
		union wait      wstatus;
		struct pending *curr = pend_head, *prev = NULL;

		/* for each entry, call waitpid */
		while (curr) {
			int done = 0;
			if ( waitpid(curr->child, &wstatus, WNOHANG) ) {
				done = ReapChild(wstatus,curr->newsock);
			}
			if (done) {
				if (prev) {
					prev->next = curr->next;
					free(curr);
					curr = prev->next;
				} else {
					pend_head = curr->next;
					free(curr);
					curr = pend_head;
				}
			} else {
				prev = curr;
				curr = curr->next;
			}
		}

		for (i = 0; i < nfds; i++) {
			if (iob[i].wclose) {
				FD_SET(i, &winit);
#ifdef TRACE_PACKETS
fprintf(stderr,"set %d winit due to close\n",i);
#endif
			}
		}

		timeout.tv_usec = 0;
		timeout.tv_sec = pend_head ? 1 : 300;

		readable = rinit;
		writable = winit;
		if ((nready = select(nfds, &readable, &writable, 0, &timeout)) == -1) {
			if (errno == EINTR)
				continue;
			perror("select");
			exit(1);
		}
		if (nready == 0) {
			tcount += pend_head ? 1 : 300;
			/* 1-1/2 hour timeout */
			if (tcount > (90*60)) {
				fprintf(stderr, "connections timed out\n");
				exit(1);
			}
		} else {
			/* reset timeout counter if there is some activity */
			tcount = 0;
		}

		for (i = 0; nready && (i < nfds); i++) {
			/* loop through descriptors */
			if (FD_ISSET(i, &writable)) {
				int nbr, leftover, o;
				o = iob[i].opposite;
				nready--;
				if (!iob[i].bytes && !iob[i].wclose) {
					fprintf(stderr,"nothing to write on %d!\n",i);
					continue;
				}
				nbr = write(i, iob[i].buffer, iob[i].bytes);
#ifdef TRACE_PACKETS
fprintf(stderr,"put %d: %d of %d bytes\n",i,nbr,iob[i].bytes);
#endif
				if (nbr == -1) {
					NotePipeError("write",i,o);
					continue;
				} else {
					leftover = iob[i].bytes - nbr;
					if ((iob[i].bytes = leftover)) {
						bcopy(iob[i].buffer + nbr,
							iob[i].buffer, leftover);
					} else {
						if (iob[i].wclose) {
							/* close after flushing */
							FD_CLR(i, &rinit);
#ifdef TRACE_PACKETS
fprintf(stderr,"clr %d rinit after flushing\n",i);
#endif
							close(i);
							iob[i].opposite = -1;
							iob[i].wclose = 0;
						}
						FD_CLR(i, &winit);
#ifdef TRACE_PACKETS
fprintf(stderr,"clr %d winit due to write\n",i);
#endif
					}
					if (o != -1) {
						/* since we wrote some, go look for more */
						FD_SET(o, &rinit);
#ifdef TRACE_PACKETS
fprintf(stderr,"set %d rinit due to write\n",o);
#endif
					}
				}
			}
			if (FD_ISSET(i, &readable)) {
				nready--;
				/* something is readable */
				if (i == 0) {
					exit(0);
				}
				if (i == lsock) {
					struct sockaddr_in dummy;
					int newsock, i, dummylen = sizeof(dummy);
					struct pending *newest;

					/* new connection ready */
					if ((newsock = accept(lsock, (struct sockaddr *) & dummy,
							 &dummylen)) < 0) {
						if (errno == EINTR)
							continue;
						perror("accept");
						exit(1);
					}
#ifndef MUST_ALLOW
	if (nokhosts == 0) {
					child_pid = fork();
					if (child_pid < 0) {
						fprintf(stderr, "Fork of child failed.\n");
						exit(1);
					}
					if (child_pid == 0) {
						/* we're the child, so create pop-up */
						CreatePopup("unknown");
					}
					/*
					 * add child to head of list of procs for which
					 * parent must check exit status
					 */
					newest = (struct pending *) malloc(sizeof(struct pending));
					newest->child = child_pid;
					newest->newsock = newsock;
					newest->next = pend_head;
					pend_head = newest;
					continue;
	}
#endif
					for (i = 0; i < nokhosts; i++) {
						if (dummy.sin_addr.s_addr == okhosts[i])
							break;
					}
					if (i == nokhosts) {
						char *cp;
						hp = gethostbyaddr((char *) &dummy.sin_addr,
								sizeof(dummy.sin_addr), dummy.sin_family);
						cp = (char*)inet_ntoa(dummy.sin_addr);
						fprintf(stderr, "bad host connect from %s (%s)\n",
								hp->h_name, cp);
						close(newsock);
						continue;
					}
					/*
					 * after check of allowed hosts,
					 * create popup on destination
					 */
					child_pid = fork();
					if (child_pid < 0) {
						fprintf(stderr, "Fork of child failed.\n");
						exit(1);
					}
					if (child_pid == 0) {
						/* we're the child, so create pop-up */
						CreatePopup(okhost_names[i]);
					}
					/*
					 * add child to head of list of procs for which
					 * parent must check exit status
					 */
					newest = (struct pending *) malloc(sizeof(struct pending));
					newest->child = child_pid;
					newest->newsock = newsock;
					newest->next = pend_head;
					pend_head = newest;
				} else {
					int nbr, o;
					o = iob[i].opposite;
					if (iob[o].bytes >= BUFSIZ) {
						fprintf(stderr,"no room to read!\n");
						continue;
					}
					nbr = read(i,
						iob[o].buffer + iob[o].bytes,
						BUFSIZ - iob[o].bytes);
#ifdef TRACE_PACKETS
fprintf(stderr,"get %d: %d bytes\n",i,nbr);
#endif
					if (nbr == -1) {
						NotePipeError("read",i,o);
						continue;
					} else if (nbr == 0) {
						/* end-of-file */
						FD_CLR(i, &rinit);
						FD_CLR(i, &winit);
#ifdef TRACE_PACKETS
fprintf(stderr,"clr %d both  due to eof\n",i);
#endif
						close(i);
						if (o != -1) {
							/* set close after finishing write */
							iob[o].wclose = 1;
							iob[o].opposite = -1;
							/* force a write cycle to clean up */
							FD_SET(o, &winit);
#ifdef TRACE_PACKETS
fprintf(stderr,"set %d winit due to clean-up\n",o);
#endif
						}
						iob[o].bytes = 0;
						iob[i].bytes = 0;
						iob[i].opposite = -1;
					} else {
						iob[o].bytes += nbr;
						FD_SET(o, &winit);
#ifdef TRACE_PACKETS
fprintf(stderr,"set %d winit due to read\n",o);
#endif
						if (iob[o].bytes >= BUFSIZ) {
							FD_CLR(i, &rinit);
#ifdef TRACE_PACKETS
fprintf(stderr,"clr %d winit due to buffer\n",i);
#endif
						}
					}
				}
			}
		}		/* for loop through descriptors */
	}			/* while (1) */
}

void
usage()
{
	fprintf(stderr, "usage: %s [-terse] [-display dispname] -allow host1 [host2 ...]\n", progname);
	exit(1);
}

int
ReapChild(wstatus,incoming)
union wait      wstatus;
int incoming;
{
	int exitstat, outgoing;

	if (WIFSIGNALED(wstatus)) {
		close(incoming);
		fprintf(stderr, "child died abnormally\n");
		return TRUE;
	}
	if (WIFEXITED(wstatus)) {
		exitstat = WEXITSTATUS(wstatus);
		if (exitstat == OK) {
			if ((outgoing = socket(PF_INET, SOCK_STREAM, 0)) < 0) {
				perror("socket");
				close(incoming);
				return TRUE;
			}
			if (connect(outgoing, (struct sockaddr *) & fwd2addr,
				 sizeof(fwd2addr)) < 0) {
				perror("connect");
				close(outgoing);
				close(incoming);
				return TRUE;
			}
			nfds = max(nfds, incoming + 1);
			nfds = max(nfds, outgoing + 1);

			iob[incoming].opposite = outgoing;
			iob[outgoing].opposite = incoming;

			iob[outgoing].bytes =
				iob[incoming].bytes = 0;
			FD_SET(incoming, &rinit);
			/* FD_SET(incoming, &winit); */
			FD_SET(outgoing, &rinit);
			/* FD_SET(outgoing, &winit); */
#ifdef TRACE_PACKETS
fprintf(stderr,"set %d,%d rinit due to open\n",incoming,outgoing);
#endif
		} else if (exitstat == CANCEL) {
			close(incoming);
		} else {
			fprintf(stderr, "Unknown exit status of child\n");
			close(incoming);
		}
		return TRUE;
	}
	/* funny status -- stopped? -- anyway just ignore */
	return FALSE;
}

void
NotePipeError(op,fd1,fd2)
char *op;
int fd1, fd2;
{
	if (errno != EPIPE) {
		fprintf(stderr, "fd %d:", fd1);
		perror(op);
	}
	/* asynchrony on close probs? */
	FD_CLR(fd1, &winit);
	FD_CLR(fd1, &rinit);
#ifdef TRACE_PACKETS
fprintf(stderr,"clr %d both! due to error\n",fd1);
#endif
	close(fd1);
	iob[fd1].opposite = -1;
	if (fd2 != -1) {
		FD_CLR(fd2, &rinit);
		FD_CLR(fd2, &winit);
#ifdef TRACE_PACKETS
fprintf(stderr,"clr %d both! due to error\n",fd2);
#endif
		close(fd2);
		iob[fd2].opposite = -1;
	}
}

#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

struct assign {
	int *var;
	int val;
};

#ifdef USE_MOTIF_GUI
#include <Xm/Xm.h>
#include <Xm/MessageB.h>

void
CreatePopup(from)
	char *from;
{
	int n = 0;
	Arg arg[4];
	XtAppContext app_con;
	Widget topshell, alert, help;
	char dialog_message[200];
	XmString msg_str, msg_str2, msg_str3;
	Display *dpy;
	int AlertResponse;
	struct assign   ok, ng;

	XtToolkitInitialize();
	app_con = XtCreateApplicationContext();
	dpy = XtOpenDisplay(app_con, fwd2name, progname, "Xforward",
			 NULL, 0, &n, NULL);
	topshell = XtAppCreateShell(NULL, "Xforward", applicationShellWidgetClass, dpy, NULL, NULL);
	sprintf(dialog_message, "Allow X connection from %s?", from);
	msg_str = XmStringCreateSimple(dialog_message);
	XtSetArg(arg[n], XmNmessageString, msg_str); n++;
	msg_str2 = XmStringCreateSimple("Yes");
	XtSetArg(arg[n], XmNokLabelString, msg_str2); n++;
	msg_str3 = XmStringCreateSimple("No");
	XtSetArg(arg[n], XmNcancelLabelString, msg_str3); n++;
	XtSetArg(arg[n], XmNdefaultButtonType, XmDIALOG_CANCEL_BUTTON); n++;
	alert = XmCreateWarningDialog(topshell, "alert", arg, n);
	XmStringFree(msg_str); XmStringFree(msg_str2); XmStringFree(msg_str3);
	ok.var = &AlertResponse; ok.val = OK;
	XtAddCallback(alert, XmNokCallback, alert_user_CB, (caddr_t) & ok);
	ng.var = &AlertResponse; ng.val = CANCEL;
	XtAddCallback(alert, XmNcancelCallback, alert_user_CB, (caddr_t) & ng);
	help = XmMessageBoxGetChild(alert, XmDIALOG_HELP_BUTTON);
	XtUnmanageChild(help);
	AlertResponse = -1;
	XtManageChild(alert);
	while (AlertResponse == -1) {
		XEvent event;
		XtAppNextEvent(app_con, &event);
		XtDispatchEvent(&event);
	}
	/* use value of AlertResponse as exit code */
	exit(AlertResponse);
}
#else
#include <X11/Shell.h>
#include <X11/Xaw/Dialog.h>

/* a stop-sign containing an exclaimation point */
#define bang_width 24
#define bang_height 24
static char     bang_bits[] = {
	0x80, 0xff, 0x01, 0xc0, 0xff, 0x03, 0xe0, 0xff, 0x07, 0xf0, 0xe7, 0x0f,
	0xf8, 0xc3, 0x1f, 0xfc, 0xc3, 0x3f, 0xfe, 0xc3, 0x7f, 0xff, 0xc3, 0xff,
	0xff, 0xc3, 0xff, 0xff, 0xc3, 0xff, 0xff, 0xe7, 0xff, 0xff, 0xe7, 0xff,
	0xff, 0xe7, 0xff, 0xff, 0xe7, 0xff, 0xff, 0xe7, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xfe, 0xe7, 0x7f, 0xfc, 0xc3, 0x3f, 0xf8, 0xc3, 0x1f,
	0xf0, 0xe7, 0x0f, 0xe0, 0xff, 0x07, 0xc0, 0xff, 0x03, 0x80, 0xff, 0x01};

void
CreatePopup(from)
	char *from;
{
	int n = 0;
	Arg arg[4];
	XtAppContext app_con;
	Widget topshell, alert;
	char message[200];
	Display *dpy;
	int AlertResponse;
	struct assign   ok, ng;
	Pixmap *bang;

	XtToolkitInitialize();
	app_con = XtCreateApplicationContext();
	dpy = XtOpenDisplay(app_con, fwd2name, progname, "Xforward",
			 NULL, 0, &n, NULL);
	bang = (Pixmap *) XCreateBitmapFromData(dpy,
					RootWindow(dpy, DefaultScreen(dpy)),
					bang_bits, bang_width, bang_height);
	topshell = XtAppCreateShell(NULL, "Xforward",
			 applicationShellWidgetClass, dpy, NULL, NULL);

	sprintf(message, "Allow X connection from %s?", from);
	XtSetArg(arg[n], XtNlabel, message); n++;
	XtSetArg(arg[n], XtNicon, bang); n++;

	alert = XtCreateManagedWidget("alert", dialogWidgetClass,
				 topshell, arg, n);
	ok.var = &AlertResponse; ok.val = OK;
	XawDialogAddButton(alert, "Yes", alert_user_CB, (caddr_t) & ok);
	ng.var = &AlertResponse; ng.val = CANCEL;
	XawDialogAddButton(alert, "No", alert_user_CB, (caddr_t) & ng);
	AlertResponse = -1;
	XtManageChild(alert);
	XtRealizeWidget(topshell);
	while (AlertResponse == -1) {
		XEvent event;
		XtAppNextEvent(app_con, &event);
		XtDispatchEvent(&event);
	}
	/* use value of AlertResponse as exit code */
	exit(AlertResponse);
}
#endif

/* ARGSUSED */
void
alert_user_CB(w, user_data, call_data)
	Widget w;
	struct assign  *user_data;
	caddr_t call_data;
{
	*(user_data->var) = user_data->val;
}

/* Check that we can open the destination display. */
void
CheckDisplay(dname)
	char *dname;
{
	Display *dpy;

	if ((dpy = XOpenDisplay(dname)) == NULL) {
		fprintf(stderr, "Unable to open destination display %s\n", dname);
		exit(1);
	}
	XCloseDisplay(dpy);
}
